// Copyright 2002-2003 University of Guelph, All Rights Reserved
import java.io.*;
import java.util.*;

class BM_CBJ2_NoArc extends Main {
  static {
	EXE_NAME = "BM_CBJ2_NoArc";
	Debug.DEBUG_LEVEL = Debug.DEBUG_MEDIUM;
  }

  public static HashSet currentDomain[];
  public static Hashtable nodes;


  public static void main(String[] Args) throws IOException, Exception {
	Main.main(Args);
// 	if ((Debug.DEBUG_LEVEL >= Debug.DEBUG_QUIET) && !quiet) {
// 	  System.out.println("Performing constraint propagation");
// 	}
// 	if ((Debug.DEBUG_LEVEL >= Debug.DEBUG_QUIET) && !quiet) {
// 	  System.out.println("Finished constraint propagation, starting search");
// 	}
	if (bcssp(problem.getMaxVars()) == NO_SOLUTION) {
	  if (useSolnFile) {
		solnOut.println("No Solution");
		solnOut.close();	  
	  } else {
		System.out.println("No Solution");
	  }
	} else {

	  if (!quiet) {
		problem.printSolution();
		System.out.println(String.valueOf(problem.numChecks) +
						   " checks, " + String.valueOf(problem.numNodes) + " nodes.");	   		
		if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MEDIUM) {
		  System.out.println(String.valueOf(nodes.size()) + " unique nodes.");
		}
	  } else {
		System.out.print(String.valueOf(problem.numChecks) + " " +
						 String.valueOf(problem.numNodes) + " ");
						 //						 String.valueOf(nodes.size()) + " ");	  
		if (useSolnFile) {
		  solnOut.print(problem.getSolutionString());
		  solnOut.close();
		}
	  }	
	}
  }

  static class LabelResult {
	int i;
	int l;
	boolean consistent;
	
	LabelResult(int I, int L, boolean Consistent) {
	  i = I;
	  l = L;
	  consistent = Consistent;
	}
  }
	  
  public static final int UNKNOWN = 0;
  public static final int NO_SOLUTION = -1;
  public static final int SOLUTION = 1;

  public static int mcl[][];
  public static int mbl[][];
  public static HashSet conflictSet[];

  public static int bcssp(int n) {
	boolean consistent = true;
	int status = UNKNOWN;
	int i = 1;
	currentDomain = new HashSet[problem.getMaxVars() + 1];
	mcl = new int[problem.getMaxVars() + 1][problem.getMaxDomain() + 1];
	mbl = new int[problem.getMaxVars() + 1][problem.getMaxDomain() + 1];
	conflictSet = new HashSet[problem.getMaxVars() + 1];
	
	for (int j = 0; j <= problem.getMaxVars(); j++) {
	  conflictSet[j] = new HashSet();
	}	

	currentDomain[0] = null;

	for (int j = 1; j <= problem.getMaxVars(); j++) {
	  currentDomain[j] = (HashSet)(problem.domain[j].clone());
	}

	for (int j = 0; j <= problem.getMaxVars(); j++) {
	  for (int k = 0; k <= problem.getMaxDomain(); k++) {
		mbl[j][k] = 0;
		mcl[j][k] = 0;
	  }
	}

	nodes = new Hashtable();

	int l = 1;

	while (status == UNKNOWN) {
	  if ((Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) && !quiet) {
		System.out.println("Starting new check after " +
						   String.valueOf(problem.numChecks) + 
						   " checks and " + String.valueOf(problem.numNodes) +
						   " nodes.");
	  }

	  if (consistent) {
		LabelResult result = bmCBJ2Label(i, consistent);
		i = result.i;
		l = result.l;
		consistent = result.consistent;
		//		problem.numNodes++;
	  } else {
		LabelResult result = bmCBJ2Unlabel(i, l, consistent);
		i = result.i;
		l = result.l; 
		consistent = result.consistent;
	  }
	  if ((Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) && !quiet) {
		System.out.println("Checks: " + String.valueOf(problem.numChecks) + " " + ", Nodes: " + String.valueOf(problem.numNodes));
	  }
	  
	  if (i > n) {
		status = SOLUTION;
	  } else if (i == 0) {
		status = NO_SOLUTION;
	  }	  
	}
	return status;
  }

  public static LabelResult bmCBJ2Label(int i, boolean consistent) {
	LabelResult result;

	consistent = false;
	Iterator kIterator = currentDomain[i].iterator();
	HashSet toRemove = new HashSet();

	int k = 1;

	while ((!consistent) && kIterator.hasNext()) {	 
	  problem.numNodes++;
	  k = ((Integer)(kIterator.next())).intValue();
	  consistent = mcl[i][k] >= mbl[i][k];

	  problem.solution[i] = k;

	  int h = mbl[i][k];
 	  mbl[i][k] = i;

	  //	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MEDIUM) {
	  //		nodes.put(problem.getNodeString(i, k), "");
	  //	  }

	  while(consistent && (h <= i - 1)) {
		//		problem.solution[i] = k;
		consistent = problem.check(i, h);
		mcl[i][k] = h;
		h++;
	  }
	  if (!consistent) {
		conflictSet[i].add(new Integer(mcl[i][k]));
		toRemove.add(new Integer(problem.solution[i]));
		//		mbl[i][k] = h;
	  }	  
	}
	if (consistent) {
	  result = new LabelResult(i + 1, k, consistent);
	} else {
	  result = new LabelResult(i, k, consistent);
	}

	Iterator lIterator = toRemove.iterator();
	while (lIterator.hasNext()) {	  
	  currentDomain[i].remove(lIterator.next());		
	}
	return result;
  }
  public static LabelResult bmCBJ2Unlabel(int i, int l, boolean consistent) {
	LabelResult result;
	int h = maxSet(conflictSet[i]);
	if (h > 0) {
	  conflictSet[h].addAll(conflictSet[i]);
	  conflictSet[h].remove(new Integer(h));
	}

	//	for (int m = l; m <= problem.getMaxDomain(); m++) {
	//	  mbl[i][m] = h;
	//	}
	//  	mbl[i][l] = h;

	for (int j = h + 1; j <= problem.getMaxVars(); j++) {
	  for (int m = 0; m <= problem.getMaxDomain(); m++) {	  
		if (mbl[j][m] > h) {
		  mbl[j][m] = h;
		}
	  }
	}

	for (int j = h + 1; j <= i; j++) {
	  conflictSet[j] = new HashSet();
	  currentDomain[j] = (HashSet)(problem.domain[j].clone());
	}
	if (h > 0) {
	  currentDomain[h].remove(new Integer(problem.solution[h]));
	  consistent = !(currentDomain[h].isEmpty());
	}
		
	result = new LabelResult(h, l, consistent);
	return result;
  }

  public static int maxSet(HashSet S) {
	Iterator i = S.iterator();
	int j;
	int maxJ = 0;

	while (i.hasNext()) {
	  j = ((Integer)i.next()).intValue();
	  if (j > maxJ) {
		maxJ = j;
	  }
	}
	return maxJ;
  }
}
